'''
Author:        Boning Dong
Last Modified: 7/7/2015


This script can allow users generate to two balls on left and right 
side of the PAGI world, and launch the balls into air by applying force.
These two balls will have a certain chance to collide. The coordinates
of two ball will be observed by PAGI guy's visual field, and will be
saved to a text file with all other data. Data for each iteration 
holds an individual line in the output file. The format of each line is
shown below.

xforce1,yforce1,xforce2,yforce2,x,y,[[coordinates],[coordinates]...]\n

#xforce1,yforce1 is the force applied to the ball on left side
#xforce2,yforce2 is the force applied to the ball on right side
#x,y is the coordinate of collision. If no collisoins happen, x and
	y will both be -1


The randomCollisionAnomaly function will add an additional force to
one of ball to avoid the collision and avoid the collision. This will
create a situation that violate physics law. The function will also
recored and save the data for each iteration. The format is similar
to the format above, except a key word for anomaly will be added after
the coordinate of collision.
A well-trained AI should be able to identify anomaly from normal cases.

'''

import math
import random
import socket
import thread
import time


def send(txt):
    #send data through socket
    clientsocket.send(txt+'\n')
    ans = clientsocket.recv(256).split(',')
    return ans
          

def create(name, x, y):
    #creat item in PAGI world
    send('createItem,'+name+',ball.png,'+str(x)+','+str(y)+',1,1,0,0,5')    


def applyForce(name, x, y, curx):
    #apply force to item to let it fly to given position

    #calculate the position difference
    xdif = x - curx
    ydif = y + 2.8
    objMass = 50
    
    #calculate the force
    hei = ydif/3*4
    t = math.sqrt(abs(2*hei/9.81))
    yforce = math.sqrt(abs(2*hei*9.81*objMass*objMass))
    #0.74 is the correction for air resistence
    xforce = xdif / t * objMass * (0.74  + abs(xdif)/1500. + abs(ydif)/1000.)
    
    #apply force
    send('addForceToItem,'+name+','+str(xforce)+','+str(yforce)+',0')
    
    #return the x and y force
    return xforce, yforce
 

def applyForceConst(name, x, y, curx):
    #apply force to item to let it fly to given position with constant time

    #calculate the position difference
    xdif = x - curx
    ydif = y + 2.8
    objMass = 50
    
    #calculate the force
    #make sure both ball reach the target position at the time
    hei = 15
    t = math.sqrt(abs(2*hei/9.81))
    yforce = math.sqrt(abs(2*hei*9.81*objMass*objMass))
    #0.74 is the correction for air resistence
    xforce = xdif / t * objMass * (0.74  + abs(xdif)/1500. + abs(ydif)/1000.)
    #change algorithm, if xforce is too large that causes sliding
    
    #apply force
    ans = send('addForceToItem,'+name+','+str(xforce)+','+str(yforce)+',0')
    
    #return the x and y force and the time of flight
    return xforce, yforce, t


def applyForceAngle(name, rad, spe):
    #apply force to item with given angle

    #calculate the force
    xforce = spe * math.cos(rad) * 50
    yforce = spe * math.sin(rad) * 50
    
    ans = send('addForceToItem,'+name+','+str(xforce)+','+str(yforce)+',0')
 

def find(code, mode, num = 128):
    #find the object in the visual field
    
    if not isinstance(num, int):
        num = 128        
    
    ans = send('findObj,'+code+','+mode)
    ans[-1] = ans[-1].strip()
    return ans


def recordData(num, useless):
    #record the data continuously

    global temp
    temp = []
    ans = []
    t1 = time.time();
    for j in range(num):
        #obtain the data through socket
        ans.append(find('obj1', 'P'))
        ans.append(find('obj2', 'P'))
    print time.time() - t1

    ans = [i[1:] for i in ans]

    #delete the empty data list
    for j in ans:
        if j != ['obj1'] and j != ['obj2']:
            temp.append(j)

def randomCollision1(num):
    #launch two balls to two random positions

    #start iteration
    for i in range(num):
        #create balls
        create('obj1', -18, -2)
        create('obj2', 18, -2)
        time.sleep(1)
        
        #randomly generate the positons
        x1 = random.uniform(-2, 2)
        x2 = random.uniform(-2, 2)
        y1 = random.uniform(8.5, 11.5)
        y2 = random.uniform(8.5, 11.5)
        
        #apply force
        applyForce('obj1', x1, y1, -18)
        applyForce('obj2', x2, y2, 18)
        time.sleep(5)
 

def randomCollision2(num):
    #launch two balls with a certain chance of collision

    #get global variable
    global text_file
    n = 0
    #start iteration
    for i in range(num):
        #create balls
        create('obj1', -20, -2)
        create('obj2', 20, -2)
        time.sleep(1)
        
        #initialize some variables
        x = y = -1
        fx1 = fy1 = fx2 = fy2 = 0
        
        #randomly decide conllision will happen or not
        if random.random() > 0.5:
            #if not collide
            #randomly generate two position that won't make
            #two balls collide
            x1 = random.uniform(-40, 20)
            x2 = random.uniform(-20, 40)
            y1 = random.uniform(5, 20)
            y2 = random.uniform(5, 20)
            
            #apply force
            fx1, fy1 = applyForce('obj1', x1, y1, -18)
            fx2, fy2 = applyForce('obj2', x2, y2, 18)            
        else:
            #if collide

            #randomly generate the collision position
            x = random.uniform(-10, 10)
            y = random.uniform(5, 15)
            
            #apply force
            fx1, fy1, t = applyForceConst('obj1', x, y, -18)
            fx2, fy2, t = applyForceConst('obj2', x, y, 18)
            
            #add the number of collision
            n += 1
        
        #record data
        recordData(150, True)
        
        #save data to string
        dstr = '%s,%s,%s,%s,%s,%s,' %(fx1,fy1,fx2,fy2,x,y)
        dstr += str(temp) + '\n'
        
        #write the string to file
        text_file.write(dstr)
        
        #destroy the items
        send('destroyItem,obj1')
        send('destroyItem,obj2')

    #return the number of collision
    return n

def anomalyHelp(t, b):
    global norm, n
    #randomly decide the ball will act abnormal or not
    if t != 0:
        rand = random.random()
    
        if rand > 0.5:
            #if the lift ball is anomaly
            time.sleep(t-0.1)
            #generate random angle and speed
            rad = random.uniform(1.57, 4.71)
            spe = random.uniform(10, 15)
            #apply force
            applyForceAngle('obj1', rad, spe)
            time.sleep(9-t)
        elif rand > 0:
            #if the right ball is anomaly
            time.sleep(t-0.1)
            #generate random angle and speed
            rad = random.uniform(4.71, 6.28)
            spe = random.uniform(5, 10)
            #apply force
            applyForceAngle('obj2', rad, spe)
            time.sleep(9-t)
        else:
            #if normal
            time.sleep(9)
            norm = 'normal'
            n += 1
    else:
        #if no collision
        time.sleep(9)
        norm = 'normal'

def randomCollisionAnomaly(num):
    #for a certain chance, an additional force will be added to
    #one of the balls to avoid collision, when it's still in the 
    #air

    #get global variable
    global text_file
    n = 0
    #start iteration
    for i in range(num):
        #create balls
        create('obj1', -20, -2)
        create('obj2', 20, -2)
        time.sleep(1)
        
        #initialize some variables
        x = y = -1
        fx1 = fy1 = fx2 = fy2 = 0
        
        #randomly decide conllision will happen or not
        if random.random() > 0.5:
            #if not collide
            #randomly generate two position that won't make
            #two balls collide
            x1 = random.uniform(-40, 20)
            x2 = random.uniform(-20, 40)
            y1 = random.uniform(5, 20)
            y2 = random.uniform(5, 20)
            
            #apply force
            fx1, fy1 = applyForce('obj1', x1, y1, -18)
            fx2, fy2 = applyForce('obj2', x2, y2, 18)            
        else:
            #if collide

            #randomly generate the collision position
            x = random.uniform(-10, 10)
            y = random.uniform(5, 15)
            
            #apply force
            fx1, fy1, t = applyForceConst('obj1', x, y, -18)
            fx2, fy2, t = applyForceConst('obj2', x, y, 18)
        
        norm = 'anomaly'
        #run anomal help function at another thread   
        try:
            thread.start_new_thread(anomalyHelp, (t, 0))
        except:
            print "Error: unable to start thread" 
    
        #record data
        recordData(150, 0)
    
        #save data to string
        dstr = '%s,%s,%s,%s,%s,%s,%s,' %(fx1,fy1,fx2,fy2,x,y,norm)
        dstr += str(temp) + '\n'
        
        #write string to file
        text_file.write(dstr)
    
        #destroy the items
        send('destroyItem,obj1')
        send('destroyItem,obj2')

    #return the number of collision
    return n

if __name__ == '__main__':
    #connect to local address
    IP_ADDRESS = socket.gethostbyname(socket.gethostname())
    #IP_ADDRESS = '129.161.58.120'
    clientsocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    clientsocket.connect((IP_ADDRESS, 42209))

    random.seed()
    
    #open file
    text_file = open('collision_output.txt','a')
    #initialize the temp variable 
    temp = []
    #start ball launching
    print randomCollision2(100)
    #close file
    text_file.close()
        